#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <libgen.h>
#include <iostream>
#include <unistd.h>
#include <fstream>



#include "SATinstance.h"
#include "SATmodel.h"
#include "BinSolver.h"

#include "global.h"

using namespace std;

Stopwatch gSW;
double preTime;

int  OrigNumVars, OrigNumClauses;

int gTimeOut;
// int numFeats=64;
int getTimeOut(void)
{
	char * value = getenv("SATTIMEOUT");

	if (value == NULL)
		return(2419200); // -- 4 weeks

	return atoi(value);
}


int main(int argc, char** argv)
{

	char *filename;
	char outfile[512];
	double thisAlgScore;
	int featureOK = 1;
	ifstream fin;
	string linefoo;
	int featNumber;
	double *rawFeat;
	char** featnames;
	int takeflag;
	long seed;
	char* strseed;

	int returnVal=0;
	SATinstance* prob=NULL;

// get current path from argv[0]
	string fullPath(argv[0]);
	size_t lastSlashPos=fullPath.find_last_of('/');
	string myownpath=fullPath.substr(0,lastSlashPos);
	if (lastSlashPos==-1){
		myownpath=".";
	}
	const char* mypath=myownpath.c_str();


//   char *dir = dirname(argv[0]);

	if (argc < 2)
	{
		printf ("usage: satzilla <instance filename> [random seed]\n");
		exit(1);
	}
	else if(argc == 3)
	{
		seed = atol(argv[2]);
		strseed=argv[2];
	}
	else
	{
		seed=(long)time(NULL);
		strseed = new char[64];
		sprintf(strseed, "%ld", seed);
	}

	filename = argv[1];
	sprintf(outfile, "%s", "/tmp");
	strcat(outfile, "/outputXXXXXX");
	mkstemp(outfile);

// -- start bookkeeping
	gTimeOut = getTimeOut();
	gSW.Start();
	BuildSolvers(strseed, outfile);


//========================================================================
// pre-solving 
	printf("c run presolver %s for %d seconds ... \n",  binSolvers[presolver1-1]->name, pretime1);
	returnVal = binSolvers[presolver1-1] -> execute(filename, pretime1, mypath);
	printf("c %s returned code %d.\n", binSolvers[presolver1-1]->name, returnVal);
	fin.open(binSolvers[presolver1-1]->outFileName);
	if (fin.is_open()){
		while (fin){
			getline(fin, linefoo);
			if(linefoo.find("s SATISFIABLE")==0){
				cout << linefoo << endl;
				returnVal = 10;
			}
			if(linefoo.find("s UNSATISFIABLE")==0){
				cout << linefoo << endl;
				returnVal = 20;
			}
			if(linefoo.find("v")==0){
                                cout << linefoo << endl;
			}
		}
		fin.close();
		remove(binSolvers[presolver1-1]->outFileName);
		binSolvers[presolver1-1]->cleanup();
		if (returnVal == 10 | returnVal == 20){
			printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
                        return returnVal;
		}
	}
	if (pretime2>0){
		printf("c run presolver %s for %d seconds ... \n",  binSolvers[presolver2-1]->name, pretime2);
		returnVal = binSolvers[presolver2-1] -> execute(filename, pretime2, mypath);
		printf("c %s returned code %d.\n", binSolvers[presolver2-1]->name, returnVal); 
		fin.open(binSolvers[presolver2-1]->outFileName);
		if (fin.is_open()){
			while (fin){
				getline(fin, linefoo);
				if(linefoo.find("s SATISFIABLE")==0){
					cout << linefoo << endl;
					returnVal = 10;
				}
				if(linefoo.find("s UNSATISFIABLE")==0){
					cout << linefoo << endl;
					returnVal = 20;
				}
				if(linefoo.find("v")==0){
                                         cout << linefoo << endl;
				}
			}
			fin.close();
			remove(binSolvers[presolver2-1]->outFileName);
			binSolvers[presolver2-1]->cleanup();
			if (returnVal == 10 | returnVal == 20){
				printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
                               return returnVal;
			}
		}
	}


// =====================================================================
// if presolver can not solve the instance 
// 1. predict feature computation time
	ifstream infile(filename);
	if (!infile) {
		fprintf(stderr, "c Error: Could not read from input file %s.\n", filename);
		exit(1);
	}
	char chbuf;
	char strbuf[1024];
	infile.get(chbuf);
	while (chbuf != 'p') {
		infile.ignore(1000, '\n');
		infile.get(chbuf);
		if(!infile)
		{
			fprintf(stderr, "c ERROR: Premature EOF reached in %s\n", filename);
			exit(1);
		}
	}
	infile >> strbuf; // "cnf"
	infile >> OrigNumVars >> OrigNumClauses;
	printf("c Orignal number of varibales is %d, number of clauses is %d \n", OrigNumVars, OrigNumClauses);

	double* featfoo;
	featfoo =new double [2];
	featfoo[0]=(double) OrigNumVars;
	featfoo[1]= (double) OrigNumClauses;

	SATmodel* featTime_Model;
	featTime_Model = new SATmodel(FTModel, FTFeatMean, FTFeatStd, FTFeatID1, FTFeatID2, FTNumFeat);
	double pred_FT=featTime_Model->estimateRunTime(featfoo);
	printf("c predicted feature computation time is %f \n", pred_FT);
	delete [] featfoo;


//2.  Now based on predicted feature computation do what ever stuff
	int bestAlgs[NUM_ALGS];
	double bestAlgScores[NUM_ALGS];

	if (pred_FT > FEAT_CUTOFF){
		featureOK=0; // no need to computer features
	}
	else{
// Now we can compute features
//		printf("c run %s as pre-processor ... \n", binSolvers[preprocess-1]->name);
//		printf("c Input file is: %s. Output file is %s\n", filename, outfile);
		returnVal = binSolvers[preprocess-1] -> execute(filename, 240, mypath);
		if (returnVal==20) {
			printf("c This instance is solved by pre-processor with %d!\n", returnVal);
			printf("s UNSATISFIABLE\n");
			printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
			return returnVal;
		}
		preTime=gSW.TotalLap();
//		printf("c Pre-process time is %f second\n", preTime);

// check if the output file is there
		if (!outfile) {
			printf("c Error: Could not read from output file\n");
			featureOK=0;
		}
		else {
			printf("c satzilla computing base features\n");
			prob = new SATinstance(outfile, seed);  
			prob->computeFeatures();
			if (prob->getNumVals()==0 ||prob->getNumClaus()==0)
			{
				cout <<"c Instance can be solved by unit propodation alone!!!\n";
				featureOK=0;
			}
		}

//3. everything is fine. computer features
		if (featureOK & doDia){
			printf("c satzilla compute diameter ...\n");
			prob->init_diameter();
		}
		if (featureOK){
			printf("c satzilla doing clause learning\n");
			prob->cl_prob(outfile, mypath);
			printf("c satzilla doing survey propogation probe\n");
			prob->sp();
		}
		if(featureOK){
			printf("c satzilla doing local search probe\n");
			prob->localSearchProbe(false);
		}
		if(featureOK & doLobo){
			printf("c satzilla doing lobjois search-space size estimate\n");
			prob->lobjoisProbe(false);
		}

//===================================================================
// do some checking
//assert(numFeats == prob->getNumFeatures());
//printf("c numFeats = %d, getNumFeatures = %d", numFeats, prob->getNumFeatures());
// classification, compute classify sore of this instance
               if (featureOK){
		featNumber = prob->getNumFeatures();
		rawFeat = prob->getFeatureVals();
		featnames= prob->getFeatureNames();
                remove(outfile);
                }

                cout << "c features are:";
		for (int i=0; i<featNumber; i++){
			cout << rawFeat[i] << " ";
//			if (rawFeat[i] == -1){
//				cout<<"Error!!! " << featnames[i]<<endl;
//				featureOK =0;
//				break;
//			}
		} 
		cout << endl;

	}

//==============================================================
// Now let predict score function
	if (featureOK==1)
	{
		for (int i=0; i<NUM_ALGS; i++)
		{
			takeflag=0;
			for (int t=0; t<subsetSize; t++){
				if(i==(solverSubset[t]-1))
					takeflag=1;
			}
			if (takeflag==1)
				thisAlgScore = (binSolvers[i]->models)->estimateRunTime(prob->getFeatureVals());
			else
				thisAlgScore = -10000;

			int j=0;
			while (j<i && thisAlgScore < bestAlgScores[j]) j++;
			for (int k=i; k>j; k--)
			{
				bestAlgs[k] = bestAlgs[k-1];
				bestAlgScores[k] = bestAlgScores[k-1];
			}
			bestAlgs[j]=i;
			bestAlgScores[j]=thisAlgScore;
		}
		printf("c satzilla's ranking:\n");
		for (int i=0; i<NUM_ALGS; i++)
		{
			printf("c %d) %s: %f\n", i+1, binSolvers[bestAlgs[i]]->name, bestAlgScores[i]);
		}
	}
	else {
		printf("c Satzilla is runing default solver ... \n");
		for (int i=0; i<NUM_ALGS; i++)
			bestAlgs[i]=i;
// we fine two best one incase one of them fails
// Do not forget -1 Dog!!!
		bestAlgs[0]=top1_DEFAULT-1;
		bestAlgs[1]=top2_DEFAULT-1;
	}


//===================================================================
// Now, we have the order. Let's run the thing
	int i=0;
	while(true)
	{
// let's tun it until someone cut it off
/*
		if(gSW.TotalLap() > gTimeOut)
		{
		printf("c Satzilla Timed out!\n");
		printf("s UNKNOWN\n");
		returnVal = 1;
		break;
	}
*/
		if(i==NUM_ALGS)
		{
			printf("c ERROR: All Failed! \n");
			returnVal = 17;
			printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
			return returnVal;
		}
// if it is a local search solver than set a cut off time for local search
		printf("c satzilla running %s ... \n", binSolvers[bestAlgs[i]]->name);
		fflush(stdout);
		returnVal = binSolvers[bestAlgs[i]] -> execute(filename, -1,mypath);
		printf("c %s returned code %d.\n", binSolvers[bestAlgs[i]]->name, returnVal);
		fin.open(binSolvers[bestAlgs[i]]->outFileName);
		if (fin.is_open()){
			while (fin){
				getline(fin, linefoo);
				if(linefoo.find("s SATISFIABLE")==0){
					cout << linefoo << endl;
					returnVal = 10;
				}
				if(linefoo.find("s UNSATISFIABLE")==0){
					cout << linefoo << endl;
					returnVal = 20;
				}
				if(linefoo.find("v")==0){
                                        cout << linefoo << endl;
				}
			}
			fin.close();
			remove(binSolvers[bestAlgs[i]]->outFileName); 
			binSolvers[bestAlgs[i]]->cleanup();   

			if (returnVal == 10 | returnVal == 20){
				printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
				remove(outfile);
			        return returnVal;
			}
		}
		i++;
	}
	return 0;
}

