#include "SATinstance.h"
#include <cstdlib>
#include <cstdio>
#include <fstream>
#include <iostream>
#include <string.h>
#include "global.h"
using namespace std;


Stopwatch gSW;
int gTimeOut;
double preTime;
int  OrigNumVars, OrigNumClauses;

// int numFeats=64;
int getTimeOut(void)
{
  char * value = getenv("SATTIMEOUT");

  if (value == NULL)
    return(2419200); // -- 4 weeks

  return atoi(value);
}

int main(int argc, char** argv)
{
  if(argc < 2)
    {
      cerr << "Usage: features {  [-all] | [-base] |[-sp] [-Dia] [-Cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
      return 1;
    }

  bool doBase=false;
  bool doUnitProbe=false;
  bool doLSProbe=false;
  bool doCl=false;
  bool doDia=false;
  bool doSp=false;
  bool doLobjois = false;
  
  int nextParam=1;
  

  if(strcmp(argv[nextParam],"-all")==0)
    {
      doBase=true;
      doUnitProbe=true;
      doLSProbe=true;
      doLobjois=true;
      doCl=true;
      doDia=true;
      doSp=true;
      nextParam++;
      if(nextParam==argc)
	{
	  cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	  return 1;
	}
    }
  else {
    if(strcmp(argv[nextParam],"-base")==0)
      {
	doBase=true;
	nextParam++;
	if(nextParam==argc)
	  {
	    cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	    return 1;
	  }
      }


    if(strcmp(argv[nextParam],"-unit")==0)
      {
	doUnitProbe=true;
	nextParam++;
	if(nextParam==argc)
	  {
	  cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	    return 1;
	  }
      }

     if(strcmp(argv[nextParam],"-sp")==0)
      {
	doSp=true;
	nextParam++;
	if(nextParam==argc)
	  {
	  cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	    return 1;
	  }
      }


   if(strcmp(argv[nextParam],"-dia")==0)
      {
	doDia=true;
	nextParam++;
	if(nextParam==argc)
	  {
	   cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	    return 1;
	  }
      }


    if(strcmp(argv[nextParam],"-cl")==0)
      {
	doCl=true;
	nextParam++;
	if(nextParam==argc)
	  {
	   cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	    return 1;
	  }
      }

    if(strcmp(argv[nextParam],"-ls")==0)
      {
	doLSProbe=true;
	nextParam++;
	if(nextParam==argc)
	  {
	  cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	    return 1;
	  }
      }

    if(strcmp(argv[nextParam],"-lobjois")==0)
      {
	doLobjois=true;
	nextParam++;
	if(nextParam==argc)
	  {
	  cerr << "Usage: features { [-all] | [-base] [-sp] [-dia] [-cl] [-unit] [-ls] [-lobjois] } infile [outfile]" << endl;
	    return 1;
	  }
      }
  }
  bool letsgo=true;
  
  if( !( doBase || doSp || doCl || doDia || doLobjois || doLSProbe || doUnitProbe) )
    doBase=true;


  char *filename;
  char outfile[512];      
  filename=argv[nextParam++];
   
  sprintf(outfile, "%s", "/tmp");
  strcat(outfile, "/outputXXXXXX");
  mkstemp(outfile);
   
  gTimeOut = getTimeOut();
  BuildSolvers("123456", outfile);
  gSW.Start();

  // before doing anything, count number of variables and clauses
  ifstream infile(filename);
  if (!infile) {
    fprintf(stderr, "c Error: Could not read from input file %s.\n", filename);
    exit(1);
  }
  char chbuf;
  char strbuf[1024];
  infile.get(chbuf);
  while (chbuf != 'p') {
    infile.ignore(1000, '\n');
    infile.get(chbuf);
    if(!infile)
      {
    	fprintf(stderr, "c ERROR: Premature EOF reached in %s\n", filename);
	    exit(1);
      }
  }
  infile >> strbuf; // "cnf"
  infile >> OrigNumVars >> OrigNumClauses;
  printf("c Orignal number of varibales is %d, number of clauses is %d \n", OrigNumVars, OrigNumClauses);


  printf("c run SatELite as pre-processor ... \n");
  int returnVal;
  printf("c Input file is: %s. Output file is %s\n", filename, outfile);
  returnVal = binSolvers[4] -> execute(filename, 1200);
  if (returnVal==10 || returnVal==20) {
      printf("c This instance is solved by pre-processor with %d!\n", returnVal);
      return -1;
      }
  

  binSolvers[4]->cleanup();

  SATinstance* sat = new SATinstance(outfile);  
  preTime=gSW.TotalLap();
  printf("c Pre-process time is %f second\n", preTime);

  if(doBase && letsgo)
  { 
    sat->computeFeatures();
 //   cout <<sat->getNumVals()<<"  "<< sat->getNumClaus()<<endl;
    if (sat->getNumVals()==0 ||sat->getNumClaus()==0)
    {
        letsgo=false;
        cout <<"c Instance can be solved by unit propodation alone!!!\n";
        } 
}

  
 if (doDia && letsgo)
     sat->init_diameter();
    
 if (doCl && letsgo)
    sat->cl_prob(outfile);
    
 if (doSp && letsgo)
    sat->sp();

  if (doUnitProbe && letsgo)
    sat->unitPropProbe(false);

  if (doLSProbe && letsgo)
    sat->localSearchProbe(false);

  if (doLobjois && letsgo)
    sat->lobjoisProbe(false);

  if (letsgo)
  if(argc > nextParam )
    {
      sat->writeFeatNamesToFile(argv[nextParam]);
      sat->writeFeaturesToFile(argv[nextParam]);
    }
  else
    {
      sat->writeFeatNamesToFile(stdout);
      sat->writeFeaturesToFile(stdout);
    }
  
  delete sat;
  remove(outfile);
  
  return 0;
}
